@@ -46,7 +46,7 @@
 #include "error.h"
 #include "sdpd.h"
 #include "btio.h"
-#include "glib-helper.h"
+#include "glib-compat.h"
 
 #include "common.h"
 #include "server.h"
@@ -88,10 +88,8 @@ static gboolean security = TRUE;
 static struct network_adapter *find_adapter(GSList *list,
 					struct btd_adapter *adapter)
 {
-	GSList *l;
-
-	for (l = list; l; l = l->next) {
-		struct network_adapter *na = l->data;
+	for (; list; list = list->next) {
+		struct network_adapter *na = list->data;
 
 		if (na->adapter == adapter)
 			return na;
@@ -102,10 +100,8 @@ static struct network_adapter *find_adapter(GSList *list,
 
 static struct network_server *find_server(GSList *list, uint16_t id)
 {
-	GSList *l;
-
-	for (l = list; l; l = l->next) {
-		struct network_server *ns = l->data;
+	for (; list; list = list->next) {
+		struct network_server *ns = list->data;
 
 		if (ns->id == id)
 			return ns;
@@ -681,10 +677,7 @@ static void server_free(struct network_server *ns)
 	g_free(ns->name);
 	g_free(ns->bridge);
 
-	if (ns->sessions) {
-		g_slist_foreach(ns->sessions, (GFunc) session_free, NULL);
-		g_slist_free(ns->sessions);
-	}
+	g_slist_free_full(ns->sessions, session_free);
 
 	g_free(ns);
 }
